@@ -407,7 +407,7 @@ static void calc_gauss(double *res, int n, double r2)
 {
     double alpha = 0.5 / r2;
     double mul = exp(-alpha), mul2 = mul * mul;
-    double cur = sqrt(alpha / M_PI);
+    double cur = sqrt(alpha / ASS_PI);
 
     res[0] = cur;
     cur *= mul;
