@@ -49,7 +49,8 @@ static char *cfstr2buf(CFStringRef string)
         size_t len = CFStringGetLength(string);
         CFIndex buf_len = CFStringGetMaximumSizeForEncoding(len, encoding);
         char *buf = malloc(buf_len);
-        CFStringGetCString(string, buf, buf_len, encoding);
+        if (buf)
+            CFStringGetCString(string, buf, buf_len, encoding);
         return buf;
     }
 }
@@ -57,7 +58,7 @@ static char *cfstr2buf(CFStringRef string)
 static void destroy_font(void *priv)
 {
     CTFontDescriptorRef fontd = priv;
-    SAFE_CFRelease(fontd);
+    CFRelease(fontd);
 }
 
 static bool is_postscript_font_format(CFNumberRef cfformat)
@@ -76,8 +77,10 @@ static bool check_postscript(void *priv)
     CTFontDescriptorRef fontd = priv;
     CFNumberRef cfformat =
         CTFontDescriptorCopyAttribute(fontd, kCTFontFormatAttribute);
+    if (!cfformat)
+        return false;
     bool ret = is_postscript_font_format(cfformat);
-    SAFE_CFRelease(cfformat);
+    CFRelease(cfformat);
     return ret;
 }
 
@@ -94,7 +97,7 @@ static bool check_glyph(void *priv, uint32_t code)
         return true;
 
     bool result = CFCharacterSetIsLongCharacterMember(set, code);
-    SAFE_CFRelease(set);
+    CFRelease(set);
     return result;
 }
 
@@ -104,13 +107,11 @@ static char *get_font_file(CTFontDescriptorRef fontd)
     if (!url)
         return NULL;
     CFStringRef path = CFURLCopyFileSystemPath(url, kCFURLPOSIXPathStyle);
-    if (!path) {
-        SAFE_CFRelease(url);
+    CFRelease(url);
+    if (!path)
         return NULL;
-    }
     char *buffer = cfstr2buf(path);
-    SAFE_CFRelease(path);
-    SAFE_CFRelease(url);
+    CFRelease(path);
     return buffer;
 }
 
@@ -120,127 +121,170 @@ static char *get_name(CTFontDescriptorRef fontd, CFStringRef attr)
     CFStringRef name = CTFontDescriptorCopyAttribute(fontd, attr);
     if (name) {
         ret = cfstr2buf(name);
-        SAFE_CFRelease(name);
+        CFRelease(name);
     }
     return ret;
 }
 
-static void process_descriptors(ASS_Library *lib, ASS_FontProvider *provider,
-                                CFArrayRef fontsd)
+static bool get_font_info_ct(CTFontDescriptorRef fontd,
+                             char **path_out,
+                             ASS_FontProviderMetaData *info)
+{
+    char *path = get_font_file(fontd);
+    *path_out = path;
+    if (!path || !*path) {
+        // skip the font if the URL field in the font descriptor is empty
+        return false;
+    }
+
+    char *ps_name = get_name(fontd, kCTFontNameAttribute);
+    info->postscript_name = ps_name;
+    if (!ps_name)
+        return false;
+
+    char *family_name = get_name(fontd, kCTFontFamilyNameAttribute);
+    info->extended_family = family_name;
+    if (!family_name)
+        return false;
+
+    return true;
+}
+
+static void process_descriptors(ASS_FontProvider *provider, CFArrayRef fontsd)
 {
     if (!fontsd)
         return;
 
-    FT_Library ftlib;
-    if (FT_Init_FreeType(&ftlib)) {
-        ass_msg(lib, MSGL_WARN, "Failed to create FT lib");
-        return;
-    }
-
     for (int i = 0; i < CFArrayGetCount(fontsd); i++) {
         ASS_FontProviderMetaData meta = {0};
         CTFontDescriptorRef fontd = CFArrayGetValueAtIndex(fontsd, i);
         int index = -1;
 
-        char *path = get_font_file(fontd);
-        if (!path || strcmp("", path) == 0) {
-            // skip the font if the URL field in the font descriptor is empty
-            free(path);
-            continue;
-        }
-
-        char *ps_name = get_name(fontd, kCTFontNameAttribute);
-
-        if (ass_get_font_info(lib, ftlib, path, ps_name, -1, &meta)) {
+        char *path = NULL;
+        if (get_font_info_ct(fontd, &path, &meta)) {
             CFRetain(fontd);
             ass_font_provider_add_font(provider, &meta, path, index, (void*)fontd);
         }
 
-        for (int j = 0; j < meta.n_family; j++)
-            free(meta.families[j]);
-
-        for (int j = 0; j < meta.n_fullname; j++)
-            free(meta.fullnames[j]);
-
-        free(meta.families);
-        free(meta.fullnames);
-
         free(meta.postscript_name);
+        free(meta.extended_family);
 
-        free(ps_name);
         free(path);
     }
-
-    FT_Done_FreeType(ftlib);
 }
 
-static void match_fonts(ASS_Library *lib, ASS_FontProvider *provider,
+static void match_fonts(void *priv, ASS_Library *lib, ASS_FontProvider *provider,
                         char *name)
 {
+    CFStringRef cfname =
+        CFStringCreateWithCString(NULL, name, kCFStringEncodingUTF8);
+    if (!cfname)
+        return;
+
     enum { attributes_n = 3 };
-    CTFontDescriptorRef ctdescrs[attributes_n];
-    CFMutableDictionaryRef cfattrs[attributes_n];
+    CTFontDescriptorRef ctdescrs[attributes_n] = {0};
     CFStringRef attributes[attributes_n] = {
         kCTFontFamilyNameAttribute,
         kCTFontDisplayNameAttribute,
         kCTFontNameAttribute,
     };
 
-    CFStringRef cfname =
-        CFStringCreateWithCString(NULL, name, kCFStringEncodingUTF8);
+    CFArrayRef descriptors = NULL;
+    CTFontCollectionRef ctcoll = NULL;
+    CFArrayRef fontsd = NULL;
 
     for (int i = 0; i < attributes_n; i++) {
-        cfattrs[i] = CFDictionaryCreateMutable(NULL, 0, 0, 0);
-        CFDictionaryAddValue(cfattrs[i], attributes[i], cfname);
-        ctdescrs[i] = CTFontDescriptorCreateWithAttributes(cfattrs[i]);
+        CFDictionaryRef cfattrs =
+            CFDictionaryCreate(NULL,
+                (const void **)&attributes[i],
+                (const void **)&cfname,
+                1, NULL, NULL);
+        if (!cfattrs)
+            goto cleanup;
+        ctdescrs[i] = CTFontDescriptorCreateWithAttributes(cfattrs);
+        CFRelease(cfattrs);
+        if (!ctdescrs[i])
+            goto cleanup;
     }
 
-    CFArrayRef descriptors =
+    descriptors =
         CFArrayCreate(NULL, (const void **)&ctdescrs, attributes_n, NULL);
+    if (!descriptors)
+        goto cleanup;
 
-    CTFontCollectionRef ctcoll =
-        CTFontCollectionCreateWithFontDescriptors(descriptors, 0);
+    const int nonzero = 1;
+    CFNumberRef cfnonzero = CFNumberCreate(NULL, kCFNumberIntType, &nonzero);
+    if (!cfnonzero)
+        goto cleanup;
+    CFDictionaryRef options =
+        CFDictionaryCreate(NULL,
+            (const void **)&kCTFontCollectionRemoveDuplicatesOption,
+            (const void **)&cfnonzero,
+            1, NULL, NULL);
+    CFRelease(cfnonzero);
+    if (!options)
+        goto cleanup;
 
-    CFArrayRef fontsd =
-        CTFontCollectionCreateMatchingFontDescriptors(ctcoll);
+    ctcoll = CTFontCollectionCreateWithFontDescriptors(descriptors, options);
+    CFRelease(options);
+    if (!ctcoll)
+        goto cleanup;
 
-    process_descriptors(lib, provider, fontsd);
+    fontsd = CTFontCollectionCreateMatchingFontDescriptors(ctcoll);
+    if (!fontsd)
+        goto cleanup;
 
+    process_descriptors(provider, fontsd);
+
+cleanup:
     SAFE_CFRelease(fontsd);
     SAFE_CFRelease(ctcoll);
 
-    for (int i = 0; i < attributes_n; i++) {
-        SAFE_CFRelease(cfattrs[i]);
+    for (int i = 0; i < attributes_n; i++)
         SAFE_CFRelease(ctdescrs[i]);
-    }
 
     SAFE_CFRelease(descriptors);
-    SAFE_CFRelease(cfname);
+    CFRelease(cfname);
 }
 
-static char *get_fallback(void *priv, const char *family, uint32_t codepoint)
+static char *get_fallback(void *priv, ASS_Library *lib,
+                          const char *family, uint32_t codepoint)
 {
     CFStringRef name = CFStringCreateWithBytes(
         0, (UInt8 *)family, strlen(family), kCFStringEncodingUTF8, false);
+    if (!name)
+        return NULL;
+
     CTFontRef font = CTFontCreateWithName(name, 0, NULL);
+    CFRelease(name);
+    if (!font)
+        return NULL;
+
     uint32_t codepointle = OSSwapHostToLittleInt32(codepoint);
+
     CFStringRef r = CFStringCreateWithBytes(
         0, (UInt8*)&codepointle, sizeof(codepointle),
         kCFStringEncodingUTF32LE, false);
+    if (!r) {
+        CFRelease(font);
+        return NULL;
+    }
+
     CTFontRef fb = CTFontCreateForString(font, r, CFRangeMake(0, 1));
-    CFNumberRef cfformat = CTFontCopyAttribute(fb, kCTFontFormatAttribute);
-    CFStringRef cfname = is_postscript_font_format(cfformat) ?
-        CTFontCopyPostScriptName(fb) : CTFontCopyFullName(fb);
-    char *res_name = cfstr2buf(cfname);
+    CFRelease(font);
+    CFRelease(r);
+    if (!fb)
+        return NULL;
 
-    SAFE_CFRelease(name);
-    SAFE_CFRelease(font);
-    SAFE_CFRelease(r);
-    SAFE_CFRelease(fb);
-    SAFE_CFRelease(cfformat);
-    SAFE_CFRelease(cfname);
+    CFStringRef cffamily = CTFontCopyFamilyName(fb);
+    CFRelease(fb);
+    if (!cffamily)
+        return NULL;
 
-    return res_name;
+    char *res_family = cfstr2buf(cffamily);
+    CFRelease(cffamily);
+
+    return res_family;
 }
 
 static void get_substitutions(void *priv, const char *name,
@@ -261,7 +305,7 @@ static ASS_FontProviderFuncs coretext_callbacks = {
 
 ASS_FontProvider *
 ass_coretext_add_provider(ASS_Library *lib, ASS_FontSelector *selector,
-                          const char *config)
+                          const char *config, FT_Library ftlib)
 {
     return ass_font_provider_new(selector, &coretext_callbacks, NULL);
 }
