@@ -638,10 +638,10 @@ shape_harfbuzz_process_run(GlyphInfo *glyphs, hb_buffer_t *buf, int offset)
         // set position and advance
         info->skip = false;
         info->glyph_index = glyph_info[j].codepoint;
-        info->offset.x    = pos[j].x_offset * info->scale_x;
-        info->offset.y    = -pos[j].y_offset * info->scale_y;
-        info->advance.x   = pos[j].x_advance * info->scale_x;
-        info->advance.y   = -pos[j].y_advance * info->scale_y;
+        info->offset.x    = lrint(pos[j].x_offset * info->scale_x);
+        info->offset.y    = lrint(-pos[j].y_offset * info->scale_y);
+        info->advance.x   = lrint(pos[j].x_advance * info->scale_x);
+        info->advance.y   = lrint(-pos[j].y_advance * info->scale_y);
 
         // accumulate advance in the root glyph
         root->cluster_advance.x += info->advance.x;
@@ -665,7 +665,7 @@ static bool shape_harfbuzz(ASS_Shaper *shaper, GlyphInfo *glyphs, size_t len)
         glyphs[i].skip = true;
 
     for (i = 0; i < len; i++) {
-        if (glyphs[i].drawing_text) {
+        if (glyphs[i].drawing_text.str) {
             glyphs[i].skip = false;
             continue;
         }
@@ -791,6 +791,42 @@ void ass_shaper_set_kerning(ASS_Shaper *shaper, bool kern)
     shaper->features[KERN].value = kern;
 }
 
+/**
+ * \brief Determine whether HarfBuzz ignores any font-provided glyph
+ * for this Unicode codepoint and replaces it with a zero-width glyph.
+ * Matches hb_unicode_funcs_t::is_default_ignorable in hb-unicode.hh.
+ * The affected codepoints are a subset of Unicode's Default_Ignorable list.
+ */
+static inline bool is_harfbuzz_ignorable(unsigned symbol) {
+    switch (symbol >> 8) {
+        case 0x00: return symbol == 0x00AD;
+        case 0x03: return symbol == 0x034F;
+        case 0x06: return symbol == 0x061C;
+        case 0x17: return symbol >= 0x17B4 && symbol <= 0x17B5;
+        case 0x18: return symbol >= 0x180B && symbol <= 0x180E;
+        case 0x20: return (symbol >= 0x200B && symbol <= 0x200F) ||
+                          (symbol >= 0x202A && symbol <= 0x202E) ||
+                          (symbol >= 0x2060 && symbol <= 0x206F);
+        case 0xFE: return (symbol >= 0xFE00 && symbol <= 0xFE0F) ||
+                          symbol == 0xFEFF;
+        case 0xFF: return symbol >= 0xFFF0 && symbol <= 0xFFF8;
+        case 0x1D1: return symbol >= 0x1D173 && symbol <= 0x1D17A;
+        default: return symbol >= 0xE0000 && symbol <= 0xE0FFF;
+    }
+}
+
+/**
+  * \brief Remove all zero-width invisible characters from the text.
+  */
+static void ass_shaper_skip_characters(GlyphInfo *glyphs, size_t len)
+{
+    for (int i = 0; i < len; i++) {
+        if (is_harfbuzz_ignorable(glyphs[i].symbol)) {
+            glyphs[i].skip = true;
+        }
+    }
+}
+
 /**
  * \brief Find shape runs according to the event's selected fonts
  */
@@ -801,23 +837,27 @@ void ass_shaper_find_runs(ASS_Shaper *shaper, ASS_Renderer *render_priv,
     int shape_run = 0;
 
     ass_shaper_determine_script(shaper, glyphs, len);
+    ass_shaper_skip_characters(glyphs, len);
 
     // find appropriate fonts for the shape runs
     for (i = 0; i < len; i++) {
         GlyphInfo *info = glyphs + i;
-        if (!info->drawing_text) {
-            // set size and get glyph index
+        if (!info->drawing_text.str && !info->skip) {
+            // get font face and glyph index
             ass_font_get_index(render_priv->fontselect, info->font,
                     info->symbol, &info->face_index, &info->glyph_index);
         }
         if (i > 0) {
             GlyphInfo *last = glyphs + i - 1;
             if ((last->font != info->font ||
-                    last->face_index != info->face_index ||
+                    (!info->skip &&
+                        last->face_index != info->face_index) ||
                     last->script != info->script ||
                     info->starts_new_run ||
                     last->flags != info->flags))
                 shape_run++;
+            else if (info->skip)
+                info->face_index = last->face_index;
         }
         info->shape_run_id = shape_run;
     }
@@ -864,33 +904,6 @@ void ass_shaper_set_bidi_brackets(ASS_Shaper *shaper, bool match_brackets)
 }
 #endif
 
-/**
-  * \brief Remove all zero-width invisible characters from the text.
-  * \param text_info text
-  */
-static void ass_shaper_skip_characters(TextInfo *text_info)
-{
-    int i;
-    GlyphInfo *glyphs = text_info->glyphs;
-
-    for (i = 0; i < text_info->length; i++) {
-        // Skip direction override control characters
-        if ((glyphs[i].symbol <= 0x202e && glyphs[i].symbol >= 0x202a)
-                || (glyphs[i].symbol <= 0x200f && glyphs[i].symbol >= 0x200b)
-                || (glyphs[i].symbol <= 0x206f && glyphs[i].symbol >= 0x2060)
-                || (glyphs[i].symbol <= 0xfe0f && glyphs[i].symbol >= 0xfe00)
-                || (glyphs[i].symbol <= 0xe01ef && glyphs[i].symbol >= 0xe0100)
-                || (glyphs[i].symbol <= 0x180f && glyphs[i].symbol >= 0x180b)
-                || glyphs[i].symbol == 0x061c
-                || glyphs[i].symbol == 0xfeff
-                || glyphs[i].symbol == 0x00ad
-                || glyphs[i].symbol == 0x034f) {
-            glyphs[i].symbol = 0;
-            glyphs[i].skip = true;
-        }
-    }
-}
-
 /**
  * \brief Shape an event's text. Calculates directional runs and shapes them.
  * \param text_info event's text
@@ -938,7 +951,6 @@ bool ass_shaper_shape(ASS_Shaper *shaper, TextInfo *text_info)
     switch (shaper->shaping_level) {
     case ASS_SHAPING_SIMPLE:
         shape_fribidi(shaper, glyphs, text_info->length);
-        ass_shaper_skip_characters(text_info);
         return true;
     case ASS_SHAPING_COMPLEX:
     default:
@@ -1020,6 +1032,11 @@ FriBidiStrIndex *ass_shaper_reorder(ASS_Shaper *shaper, TextInfo *text_info)
     return shaper->cmap;
 }
 
+FriBidiStrIndex *ass_shaper_get_reorder_map(ASS_Shaper *shaper)
+{
+    return shaper->cmap;
+}
+
 /**
  * \brief Resolve a Windows font charset number to a suitable base
  * direction. Generally, use LTR for compatibility with VSFilter. The
